//************************************************************************
// FileName:        PixieMAC.c
//                  IEEE 802.15.4 communications 
//
// One of the following macros should be defined in the Build Options
// #define OSC_INTERNAL  // sets the config bits for internal oscillator
// #define OSC_4MHz  	// sets the config bits 4MHz oscillator with 4 X PLL to get 16MHz 
// #define OSC_16MHz  	// sets the config bits 16MHz 
//
// One of the following macros should be defined in the Build Options
// #define CODE_PROTECT  	// code protection on
// #define NO_CODE_PROTECT 	// code protection off
//
//*************************************************************************

// includes

#include <p18cxxx.h>   

#include "MACAPI.h"
#include <delays.h> 

//******************************************************************************

// function prototypes
void TxChar(char ch);
void PutROMString(ROM char* str);
void PutROMHex(ROM char* str, BYTE len);
void PutHex(char* str, BYTE len);
void TxHex(BYTE toPrint);
BYTE TranslateHexChar( BYTE ch );
void TryStartTx( void );

//********************************************************************************/
// semaphores

typedef union _APP_FLAGS
{
    BYTE Val;
    struct _Bits
    { 
		BYTE CmdReady : 1;
		BYTE ParseError : 1;
		BYTE Scanning : 1;
		BYTE SendContinueWhenIdle : 1;
	} Bits;
}  APP_FLAGS;

APP_FLAGS AppFlags = 0b00000000;  

//********************************************************************************/

// serial buffers in and out

#pragma udata RxBuffer
BYTE RxBuff[256];
#pragma udata TxBuffer
BYTE TxBuff[256];
#pragma udata 
BYTE RxStart = 0;
BYTE RxNext = 0;
BYTE TxStart = 0;
BYTE TxNext = 0;

// Temporary buffers
BYTE cTemp;
BYTE* pTemp;
BYTE TxParamBytes;

//********************************************************************************/

typedef struct 
{
	BYTE ID;
	rom char Msg[4];
	unsigned char nArgBytes;
	BYTE MACPrimitive;
} RxMessage;

#define RxID_DMCR 0x00
#define RxID_DSMR (RxID_DMCR+1)
#define RxID_DVRR (RxID_DSMR+1)
#define RxID_DHTR (RxID_DVRR+1)
#define RxID_MDAR (RxID_DHTR+1)
#define RxID_MPGR (RxID_MDAR+1)
#define RxID_MASR (RxID_MPGR+1)
#define RxID_MASS (RxID_MASR+1)
#define RxID_MDSR (RxID_MASS+1)
#define RxID_MGTR (RxID_MDSR+1)
#define RxID_MTSR (RxID_MGTR+1)
#define RxID_MORS (RxID_MTSR+1)
#define RxID_MRSR (RxID_MORS+1)
#define RxID_MRXR (RxID_MRSR+1)
#define RxID_MSCR (RxID_MRXR+1)
#define RxID_MSYR (RxID_MSCR+1)
#define RxID_MSRR (RxID_MSYR+1)
#define RxID_MPLR (RxID_MSRR+1)
#define RxID_MSTR (RxID_MPLR+1)
#define RxID_MRTS (RxID_MSTR+1)
#define RxID_MRRX (RxID_MRTS+1)
#define RxID_MRSY (RxID_MRRX+1)
#define RxID_ENDS (RxID_MRSY+1)

#define VARIABLE_LENGTH 0xFE	

rom RxMessage RxCommand[] =   { 		{ RxID_DMCR, {'D','M','C','R'}, 0, 0 } ,   
										{ RxID_DSMR, {'D','S','M','R'}, 8, 0 } ,   
										{ RxID_DVRR, {'D','V','R','R'}, 0, 0 } ,   
										{ RxID_DHTR, {'D','H','T','R'}, 0, 0 } ,   
										{ RxID_MDAR, {'M','D','A','R'}, 27, MCPS_DATA_request } ,	// data
										{ RxID_MPGR, {'M','P','G','R'}, 27, MCPS_PURGE_request } ,	// purge
										{ RxID_MASR, {'M','A','S','R'}, 26, MLME_ASSOCIATE_request } ,	// associate
										{ RxID_MASS, {'M','A','S','S'}, 24, MLME_ASSOCIATE_response } ,	// associate
										{ RxID_MDSR, {'M','D','S','R'}, 16, MLME_DISASSOCIATE_request } ,	// disassociate
										{ RxID_MGTR, {'M','G','T','R'}, 1, MLME_GET_request } ,	// get
										{ RxID_MTSR, {'M','T','S','R'}, 0, MLME_GTS_request } ,	// gts
										{ RxID_MORS, {'M','O','R','S'}, 18, MLME_ORPHAN_response } ,	// orphan
										{ RxID_MRSR, {'M','R','S','R'}, 1, MLME_RESET_request } ,	// reset
										{ RxID_MRXR, {'M','R','X','R'}, 0, MLME_RX_ENABLE_request } ,	// rx enable
										{ RxID_MSCR, {'M','S','C','R'}, 12, MLME_SCAN_request } ,	// scan
										{ RxID_MSYR, {'M','S','Y','R'}, 0, MLME_SYNC_request } ,	// sync
										{ RxID_MSRR, {'M','S','R','R'}, 14, MLME_START_request } ,	// start
										{ RxID_MPLR, {'M','P','L','R'}, 26, MLME_POLL_request } ,	// poll
										{ RxID_MSTR, {'M','S','T','R'}, 1, MLME_SET_request } ,	// set
										{ RxID_ENDS, {'0','0','0','0'}, 0, NO_PRIMITIVE } ,	// end
									  	};  

// currently there is no actual verification that the 
// correct mailbox library has been linked in, you need to get that 
// right in the project settings
#if ( defined(PXMC) )
#define VERSION_CONFIRM 			(rom char *) "+DVRC=0B40011310362105110601 www.flexipanel.com\r\n"
#elif (defined(PXMF) )
#define VERSION_CONFIRM 			(rom char *) "+DVRC=0B40011310362105110602 www.flexipanel.com\r\n"
#elif (defined(PXMS) )
#define VERSION_CONFIRM 			(rom char *) "+DVRC=0B40011310362105110603 www.flexipanel.com\r\n"
#elif (defined(PLMC) )
#define VERSION_CONFIRM 			(rom char *) "+DVRC=0B40011310362105110604 www.flexipanel.com\r\n"
#elif (defined(PLMF) )
#define VERSION_CONFIRM 			(rom char *) "+DVRC=0B40011310362105110605 www.flexipanel.com\r\n"
#elif (defined(PLMS) )
#define VERSION_CONFIRM 			(rom char *) "+DVRC=0B40011310362105110606 www.flexipanel.com\r\n"
#else
#error "Stack / device not defined" 
#endif
#define NOT_UNDERSTOOD_INDICATION 	(rom char *) "+DERI=01\r\n"
#define MAC_NOT_SET				 	(rom char *) "+DERI=04\r\n"
#define NO_MAC_RESET				(rom char *) "+DERI=03\r\n"
#define NOT_SUPPORTED				(rom char *) "+DERI=02\r\n"
#define CRLF						(rom char *) "\r\n"
#define MSCC						(rom char *) "+MSCC="
#define MBNI						(rom char *) "+MBNI="
#define MDAC 						(rom char *) "+MDAC="
#define MRSC						(rom char *) "+MRSC="
#define MPGC						(rom char *) "+MPGC="
#define MASI						(rom char *) "+MASI="
#define MASC						(rom char *) "+MASC="
#define MDSI						(rom char *) "+MDSI="
#define MDSC						(rom char *) "+MDSC="
#define MORI						(rom char *) "+MORI="
#define MCSI						(rom char *) "+MCSI="
#define MGTC						(rom char *) "+MGTC="
#define MSTC						(rom char *) "+MSTC="
#define MSRC						(rom char *) "+MSRC="
#define MPLC						(rom char *) "+MPLC="
#define MSLI						(rom char *) "+MSLI="
#define DCTI						(rom char *) "+DCTI\r\n"
#define MDAI						(rom char *) "+MDAI="
#define PDAI						(rom char *) "+PDAI="
#define DHTC						(rom char *) "+DHTC\r\n"
#define DGMC						(rom char *) "+DGMC="
#define DSMC						(rom char *) "+DSMC=00\r\n"
#define DRYI						(rom char *) "+DRYI\r\n"

//-------------------------------------------------------------------------------------------------------

// Pin definitions
#define RTSpinT 		TRISAbits.TRISA1
#define RTSpin 			PORTAbits.RA1
#define CTSpinT 		TRISAbits.TRISA2
#define CTSpin 			PORTAbits.RA2
#define SleepPinT 		TRISBbits.TRISB4
#define SleepPin 		PORTBbits.RB4


//--------------------------------------------------------------------------------------------------------

// MAC state machine

ZIGBEE_PRIMITIVE    currentPrimitive;

//--------------------------------------------------------------------------------------------------------

// main program

void main(void) 
{
	OSCCONbits.IRCF1 = 1;  	// assuming <IDCF2:IDCF0>=100 on startup, changes to <IDCF2:IDCF0>=110 = 4MHz
	OSCTUNEbits.PLLEN = 1;  // PLL 4MHz -> 16MHz
	Delay1KTCYx( 100 ); 	// allow 25ms for clock to settle
    ADCON1 = 0x0F; 		 // Make PORTA digital I/O.
	TRISAbits.TRISA5 = 0;
	
	// initialize I/O. 
    ADCON1 = 0x0F; 		 // Make PORTA digital I/O.
    RTSpin = 1;
    RTSpinT = 0;
    CTSpinT = 1;
    SleepPinT = 1;

	// set up internal oscillator for 16MHz operation if required
#ifdef OSC_INTERNAL
	OSCCONbits.IRCF1 = 1;  	// assuming <IDCF2:IDCF0>=100 on startup, changes to <IDCF2:IDCF0>=110 = 4MHz
	OSCTUNEbits.PLLEN = 1;  // PLL 4MHz -> 16MHz
	Delay1KTCYx( 100 ); 	// allow 25ms for clock to settle
#endif

	// enable watchdog timer 
    ClrWdt(); 
    ENABLE_WDT();
 
    // Initialize UART for 115200 MHz
    BAUDCONbits.BRG16 = 1;
    TXSTA = 0x24;
    RCSTA = 0x90; // 0b10010000;
    SPBRG = 34;  
    SPBRGH = 0;
    RCIP = 1;
    RCIE = 1;
    TXIP = 0;
    TXIE = 0;
 	INTCON2bits.RBIP = 1;
    RBIE = 0;
    
    // Initialize MAC layer
    currentPrimitive = NO_PRIMITIVE; 
    MACAPIInit();
    RTSpin = 0;
	PutROMString( DRYI );
   
    while (1)
    { 
		// perform mailbox tasks
	    ClrWdt(); 
        MACAPITasks( &currentPrimitive );
        
		// restart output buffer if CTS has cleared
		TryStartTx();
		
        switch (currentPrimitive)
        {
			case PD_DATA_indication:

				PutROMString( PDAI );
				PutHex( (void *)&params.PD_DATA_indication.psduLength, 1 );
				PutHex( (void *)params.PD_DATA_indication.psdu, params.PD_DATA_indication.psduLength  );
				PutROMString( CRLF );

                MACDiscardRx();
                currentPrimitive = NO_PRIMITIVE;
                break;

			case MLME_BEACON_NOTIFY_indication: 

				PutROMString( MBNI );
				PutHex( (void *)&params, sizeof( params.MLME_BEACON_NOTIFY_indication )   );
				PutHex( (void *)params.MLME_BEACON_NOTIFY_indication.sdu, params.MLME_BEACON_NOTIFY_indication.sduLength  );
				PutROMString( CRLF );

				MACDiscardRx( );
                currentPrimitive = NO_PRIMITIVE;
                break;

			case MCPS_DATA_indication:

				PutROMString( MDAI );
				PutHex( (void *)&params, sizeof( params.MCPS_DATA_indication )   );
				PutHex( (void *)params.MCPS_DATA_indication.msdu, params.MCPS_DATA_indication.msduLength  );
				PutROMString( CRLF );

				MACDiscardRx( );
                currentPrimitive = NO_PRIMITIVE;
                break;


			case MLME_SCAN_confirm: 
				AppFlags.Bits.Scanning = 0;
				if (params.MLME_SCAN_confirm.ScanType == MAC_SCAN_ENERGY_DETECT)
				{
					TxParamBytes = 	params.MLME_SCAN_confirm.ResultListSize + sizeof( params.MLME_SCAN_confirm );
				}
				else if (params.MLME_SCAN_confirm.ScanType == MAC_SCAN_ORPHAN_SCAN)
				{
					TxParamBytes = 	sizeof( params.MLME_SCAN_confirm );
				}
				else  
				{
					TxParamBytes = 	params.MLME_SCAN_confirm.ResultListSize * sizeof ( PAN_DESCRIPTOR ) 
										+ sizeof( params.MLME_SCAN_confirm );

				}

//				if (TxParamBytes > PARAMBUFF) TxParamBytes = PARAMBUFF;				

				PutROMString( MSCC );
				PutHex( (void *)&params, sizeof( params.MLME_SCAN_confirm )  );
				PutHex( (void *)params.MLME_SCAN_confirm.EnergyDetectList, /* assumes pointer is same for ed as pan descr */
							TxParamBytes - sizeof( params.MLME_SCAN_confirm )  );
				PutROMString( CRLF );

				// free up memory
				if (params.MLME_SCAN_confirm.ScanType == MAC_SCAN_ENERGY_DETECT)
				{
                    if (params.MLME_SCAN_confirm.EnergyDetectList)
                    {
                        SRAMfree( (unsigned char *) params.MLME_SCAN_confirm.EnergyDetectList );
                    }
				}
				else if (params.MLME_SCAN_confirm.ScanType == MAC_SCAN_ORPHAN_SCAN)
				{
				}
				else  
				{
					// delete descriptor list
                    while ((BYTE *)params.MLME_SCAN_confirm.PANDescriptorList)
                    {
                        BYTE  * ptr = (BYTE *)params.MLME_SCAN_confirm.PANDescriptorList->next;
                        SRAMfree( (unsigned char *) params.MLME_SCAN_confirm.PANDescriptorList );
                        params.MLME_SCAN_confirm.PANDescriptorList = (PAN_DESCRIPTOR *)ptr;
                    }
				}

                currentPrimitive = NO_PRIMITIVE;
                break;
                
            case MCPS_DATA_confirm:
				// don't pass up if in the middle of a scan
				if (AppFlags.Bits.Scanning)
				{
	                currentPrimitive = NO_PRIMITIVE;
	                break;
				}
				AppFlags.Bits.SendContinueWhenIdle = 0;
				PutROMString( MDAC );
				TxParamBytes = sizeof( params.MCPS_DATA_confirm );
OutputPacket:
				PutHex( (void *)&params, TxParamBytes  );
				PutROMString( CRLF );
                currentPrimitive = NO_PRIMITIVE;
                break;

            case MLME_RESET_confirm:
				PutROMString( MRSC );
				TxParamBytes = sizeof( params.MLME_RESET_confirm );				
				goto OutputPacket;

			case MCPS_PURGE_confirm:
				PutROMString( MPGC );
				TxParamBytes = sizeof( params.MCPS_PURGE_confirm );
				goto OutputPacket;

			case MLME_ASSOCIATE_indication:
				PutROMString( MASI );
				TxParamBytes = sizeof( params.MLME_ASSOCIATE_indication );
				goto OutputPacket;

			case MLME_ASSOCIATE_confirm: 
				PutROMString( MASC );
				TxParamBytes = sizeof( params.MLME_ASSOCIATE_confirm );
				goto OutputPacket;

			case MLME_DISASSOCIATE_indication:
				PutROMString( MDSI );
				TxParamBytes = sizeof( params.MLME_DISASSOCIATE_indication );
				goto OutputPacket;
 
			case MLME_DISASSOCIATE_confirm:
				PutROMString( MDSC );
				TxParamBytes = sizeof( params.MLME_DISASSOCIATE_confirm );
				goto OutputPacket;

			case MLME_ORPHAN_indication:
				PutROMString( MORI );
				TxParamBytes = sizeof( params.MLME_ORPHAN_indication );
				goto OutputPacket;

			case MLME_COMM_STATUS_indication:
				PutROMString( MCSI );
				TxParamBytes = sizeof( params.MLME_COMM_STATUS_indication );
				goto OutputPacket;

			case MLME_START_confirm:
				PutROMString( MSRC );
				TxParamBytes = sizeof( params.MLME_START_confirm );
				goto OutputPacket;

			case MLME_POLL_confirm:
				PutROMString( MPLC );
				TxParamBytes = sizeof( params.MLME_POLL_confirm );
				goto OutputPacket;

			case MLME_SYNC_LOSS_indication:
				PutROMString( MSLI );
				TxParamBytes = sizeof( params.MLME_SYNC_LOSS_indication );
				goto OutputPacket;

			case MLME_SET_confirm:
				PutROMString( MSTC );
				TxParamBytes = sizeof( params.MLME_SET_confirm );
				goto OutputPacket;

			case MLME_GET_confirm:
				PutROMString( MGTC );
				PutHex( (void *)&params, 2  );
				PutHex( (void *)params.MLME_GET_confirm.MIBAttributeValue, params.MLME_GET_confirm.MIBAttributeLen  );
				PutROMString( CRLF );
                currentPrimitive = NO_PRIMITIVE;
                break;

            default:
            	break;
        }
 
 
		// Pump the stack tasks until idle
		if (currentPrimitive!=NO_PRIMITIVE)  continue;
		
		// if midway through an indirect data request, indicate next command may proceed
		if (AppFlags.Bits.SendContinueWhenIdle)
		{
			AppFlags.Bits.SendContinueWhenIdle = 0;
			PutROMString( DCTI );
			continue;
		}
		
		// process incoming message if any
		if (AppFlags.Bits.CmdReady)
		{
			// fast forward through any stray characters until reach a '+' or run out of data
			while (RxBuff[RxStart]!='+' && RxStart!=RxNext) RxStart++;
			
			// identify command
			if (RxStart!=RxNext)
			{
		        char iCommand, iArg, bError;

		        for (iCommand = 0; RxCommand[iCommand].Msg[0]!=0; iCommand++)
		        {													// &0xDF for case insensitivity
			        if (RxCommand[iCommand].Msg[0]==(RxBuff[RxStart+1]&0xDF) && 
			        	RxCommand[iCommand].Msg[1]==(RxBuff[RxStart+2]&0xDF) && 
			        	RxCommand[iCommand].Msg[2]==(RxBuff[RxStart+3]&0xDF) && 
			        	RxCommand[iCommand].Msg[3]==(RxBuff[RxStart+4]&0xDF) )  break;
			    }
			    
			    if (RxCommand[iCommand].Msg[0])  
			    {
				    // found message

					// if command is data and another transmit buffer is in use, pump the stack until freed, otherwise reserve bufer
					if ( iCommand == RxID_MDAR )
					{
						if (IsMACReady())
							MACBlockTx();
						else
							continue;
					}
				
				    // zoom forward to arguments '=' is optional
				    RxStart += 5;
				    if (RxBuff[RxStart]=='=')
				    {
					    RxStart++;
					}
					else
					{
						if (RxCommand[iCommand].nArgBytes!=0 || RxBuff[RxStart]!='\r')  goto ParseError;
					}
					
				    // Parse arguments into (&params)
				    AppFlags.Bits.ParseError = 0;
				    
					// translate fixed length arguments
					for (iArg = 0; iArg<RxCommand[iCommand].nArgBytes; iArg++)
					{
						((BYTE*)(&params))[iArg] = (TranslateHexChar( RxBuff[RxStart] ) << 4) + TranslateHexChar( RxBuff[RxStart+1] );
						RxStart+=2;
						if (AppFlags.Bits.ParseError)  goto ParseError;
					}
					
					// translate variable length arguments
					cTemp = 0;
				    if (iCommand==RxID_MDAR )
				    {
					    pTemp = pTxData;
					    cTemp = (&params)->MCPS_DATA_request.msduLength;
					}
				    if (iCommand==RxID_MSTR )
				    {	// pull off as many bytes as we're given and store at beginning of RxBuff itself, snice it's otherwise unused right now 
					    cTemp = 0;
						pTemp = &RxBuff[RxStart];
						while (*pTemp != '\r' && cTemp<0x10) 
						{
							cTemp++;
							pTemp+=2;
						}
					    pTemp = RxBuff;
					    (&params)->MLME_SET_request.MIBAttributeValue = RxBuff;
					}
					
					// parse variable length characters
				    for ( ; cTemp; cTemp--)
				    {
						*pTemp = (TranslateHexChar( RxBuff[RxStart] ) << 4) + TranslateHexChar( RxBuff[RxStart+1] );
						RxStart+=2;
						if (AppFlags.Bits.ParseError)  goto ParseError;
						pTemp++;
					}
					
					// check that was the end of the message
	        		if (RxBuff[RxStart]!='\r') goto ParseError;
					
					// message parsed correctly, process it
					if (iCommand==RxID_DVRR) 
					{
						if (!IsMACAddressValid()) 
							PutROMString( MAC_NOT_SET );
						else
							PutROMString( VERSION_CONFIRM );
					}
					else if (iCommand==RxID_DHTR) 
					{
						bNoTimeout = 1;
						PutROMString( DHTC );
					}
					else if (iCommand==RxID_DSMR) 
					{
						if (IsMACAddressValid())
							PutROMString( NO_MAC_RESET );
						else
						{
							SetMACAddress( (&params) );
							((BYTE*)(&params))[0] = 1;
							PutROMString( DSMC );
							while (TXIE || TxNext!=TxStart || !TRMT) ClrWdt();
							Reset();
						}
					}
					else if (iCommand==RxID_DMCR) 
					{
						PutROMString( DGMC );
						PutROMHex((ROM char*) &macLongAddr, sizeof(macLongAddr));
						PutROMString( CRLF );
					}
					else if (iCommand==RxID_MRTS || iCommand==RxID_MRRX || iCommand==RxID_MRSY)
					{
						PutROMString( NOT_SUPPORTED );
						PutROMHex((ROM char*) &macLongAddr, sizeof(macLongAddr));
						PutROMString( CRLF );
					}
					else
					{
						// remaining messages are handled by the stack, but check MAC address has been set
						if (!IsMACAddressValid()) 
							PutROMString( MAC_NOT_SET );
						else
						{
							currentPrimitive = RxCommand[iCommand].MACPrimitive;
							if (currentPrimitive==MLME_SCAN_request) AppFlags.Bits.Scanning = 1;
							if (currentPrimitive==MCPS_DATA_request &&
							    (&params)->MCPS_DATA_request.TxOptions.bits.indirect_transmission) 	AppFlags.Bits.SendContinueWhenIdle = 1;
							if (currentPrimitive==MLME_ORPHAN_response || currentPrimitive==MLME_ASSOCIATE_response) 	AppFlags.Bits.SendContinueWhenIdle = 1;
						}
					}
				}
				else
				{
ParseError:
					// error - denerate "did not understand" message
					PutROMString( NOT_UNDERSTOOD_INDICATION );
				}				
			}
			
			// discard buffer
			AppFlags.Bits.CmdReady = 0;
			RxStart = RxNext;
		}
		
#if (defined( PXMS ) || defined( PLMS ))  // only PXMS / PLMS devices are allowed to sleep

        // If we don't have to execute a primitive, see if we need to request data from
        // our parent, or if we can go to sleep.
        if (currentPrimitive == NO_PRIMITIVE && SleepPin==1)
        {
	        if (StackCanSleep())
			{
				if (TxStart==TxNext)  // check finished outputting data
				{
                  	MACDisable();
				    RTSpin = 1;
    	            RBIF = 0;
    	            RBIE = 1;
					DISABLE_WDT();

					Sleep();
              		Nop();

					ClrWdt(); 
					ENABLE_WDT();
				
					MACEnable();
				    RTSpin = 0;
				}
			}
      	}
#endif

   }  // end of while (1) loop
}


/*******************************************************************************
Serial tx and rx routines
*******************************************************************************/

void TryStartTx( void )
{
	char ch = CTSpin;
	ch = TXIE;
		if (TxStart!=TxNext && CTSpin==0 && !TXIE)
	{
		// more to transmit
		TXREG = TxBuff[TxStart];
		TxStart++;
		TXIE = 1;
	}
}

void TxChar(char ch)
{
    TxBuff[TxNext] = ch;
    TxNext++;  // no overflow checking
        
	TryStartTx();
}

void PutROMString(ROM char* str)
{
    BYTE c;
    while( c = *str++ )  TxChar( c );
}

void PutROMHex(ROM char* str, BYTE len)
{
    BYTE c;
    while( len )  
    {
	    c = *str++;
	    TxHex( c );
	    len--;
	}
}

void PutHex(char* str, BYTE len)
{
    BYTE c;
    while( len )  
    {
	    c = *str++;
	    TxHex( c );
	    len--;
	}
}

ROM unsigned char CharacterArray[]={'0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F'};
void TxHex(BYTE toPrint)
{
    BYTE PRINT_VAR;
    PRINT_VAR = toPrint;
    toPrint = (toPrint>>4)&0x0F;
    TxChar(CharacterArray[toPrint]);
    toPrint = (PRINT_VAR)&0x0F;
    TxChar(CharacterArray[toPrint]);
    return;
}

BYTE TranslateHexChar( BYTE ch )
{
	// Convert from hex character into byte value form 0 to 15
	ch = ch - '0';
	if (ch > 9) ch -= 0x07;  // convert ABCDEF
	if (ch > 0x0F) ch -= 0x20;  // convert abcdef
	if (ch > 0x0F) AppFlags.Bits.ParseError = 1;
	return ch;
}


/*******************************************************************************
User Interrupt Handlers

The stack uses some interrupts at low priority for its internal processing.  Once it is done
checking for its interrupts, the stack calls UserInterruptHandler function to allow for any
additional interrupt processing.  Interrupts should not be disabled unless the sleep state 
has been entered.  

PriorityUserInterruptHandler must be for very fast, defined duration operations only

*******************************************************************************/


#pragma interrupt PriorityUserInterruptHandler
void PriorityUserInterruptHandler(void)
{
	static BYTE InChar;
	if (RCIF)
	{
		InChar = RCREG;  
		if ( OERR || FERR ) // overflow or framing error, reset UART
		{
			CREN = 0;
			CREN = 1;
		}
		else
		{
			RxBuff[RxNext] = InChar;
			RxNext++;
			if (InChar=='\r') 
			{
				// raise semaphore that a command is ready to process.
				AppFlags.Bits.CmdReady = 1;
			}
		}
	}
	// clear if interrupt on change
	else if ( RBIE && RBIF )  
	{
        RBIE = 0;
        RBIF = 0;
	}
}


void UserInterruptHandler(void)
{
	// transmission is lower priority than reception
	if (TXIE && TXIF)
	{
		if (TxStart!=TxNext && CTSpin==0)
		{
			// more to transmit
			TXREG = TxBuff[TxStart];
			TxStart++;
		}
		else
		{
			TXIE = 0;
		}
	}
} 


/*******************************************************************************

MACAPIHook

This function allows a sneak look at (but not touch!) the state of the stack 
Below are two example if you #define DIAGNOSTICS
- Reporting the state of the stack
- Extracting RSSI and LQI from received packets
You will have to examine the Microchip ZigBee stack source for details of how they work

********************************************************************************/

#include "MACTasks.h"

extern PARAMS           		params;
extern volatile ZIGBEE_STATUS  	ZigBeeStatus;
static unsigned char ZigBeeState = 0x00;

rom char* rom ZB_State[] = { 	"NO_PRIMITIVE", 	"PD_DATA_req", 	"PD_DATA_conf",	"PD_DATA_ind", 
							"PLME_CCA_req", "PLME_CCA_conf", "PLME_ED_req", 	"PLME_ED_conf",
							"PLME_GET_req", "PLME_GET_conf", "PLME_SET_TRX_STATE_req", 	"PLME_SET_TRX_STATE_conf",
 							"PLME_SET_req", "PLME_SET_conf", "", "", 
							"MCPS_DATA_req", 	"MCPS_DATA_conf", "MCPS_DATA_ind", "MCPS_PURGE_req", 
							"MCPS_PURGE_conf", 	"MLME_ASSOCIATE_req", "MLME_ASSOCIATE_ind", "MLME_ASSOCIATE_rsp", 
							"MLME_ASSOCIATE_conf", 	"MLME_DISASSOCIATE_req", "MLME_DISASSOCIATE_ind", "MLME_DISASSOCIATE_conf", 
							"MLME_BEACON_NOTIFY_ind", 	"MLME_GET_req",	"MLME_GET_conf",	"MLME_GTS_req",
							"MLME_GTS_conf", "MLME_GTS_ind", "MLME_ORPHAN_ind", 	"MLME_ORPHAN_rsp",
							"MLME_RESET_req", "MLME_RESET_conf", "MLME_RX_ENABLE_req", 	"MLME_RX_ENABLE_conf",
							"MLME_SCAN_req", "MLME_SCAN_conf", "MLME_COMM_STATUS_ind", 	"MLME_SET_req",
							"MLME_SET_conf", "MLME_START_req", "MLME_START_conf", 	"MLME_SYNC_req",
							"MLME_SYNC_LOSS_ind", "MLME_POLL_req" 
 };

void MACAPIHook( void )
{
#ifdef	DIAGNOSTICS
	// report stack state if changed
	if (ZigBeeState != (unsigned char) ZigBeeStatus.nextZigBeeState)
	{
        PutROMString( (rom char *)" ZigBee State:  " );
		PutROMString( ZB_State[ZigBeeState] );
       	PutROMString( (rom char *)" -> " );

		ZigBeeState = ZigBeeStatus.nextZigBeeState;
		PutROMString( ZB_State[ZigBeeState] );
       	PutROMString( (rom char *)"\r\n" );
	}	

	// sneak-read RSSI if this is a PD_DATA_indication
	if (ZigBeeStatus.nextZigBeeState==PD_DATA_indication)
	{
		// refer to CC2420 documentation for details of how these values are derived and what they actually mean
		unsigned char LQIVal, RSSIVal;
		LQIVal = (params.PD_DATA_indication.psdu[params.PD_DATA_indication.psduLength-1] & 0x7F) - 10; // 0x00 = 0%, 0x64 = 100%
		RSSIVal = params.PD_DATA_indication.psdu[params.PD_DATA_indication.psduLength-2]+ 65;  			// 00 = -110dBm, 6E = 0dBm
        PutROMString( (rom char *)" Packet Recvd:  RSSI = " );
		TxHex(LQIVal);	
        PutROMString( (rom char *)",  LQI = " );
		TxHex(RSSIVal);	
       	PutROMString( (rom char *)"\r\n" );
	}
#endif
}
